@@ -1,10 +1,9 @@
 /*
- * Copyright 2005-2007 Gentoo Foundation
+ * Copyright 2005-2012 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
- * $Header: /var/cvsroot/gentoo-projects/pax-utils/paxinc.h,v 1.10 2007/08/20 09:54:15 vapier Exp $
  *
- * Copyright 2005-2007 Ned Ludd        - <solar@gentoo.org>
- * Copyright 2005-2007 Mike Frysinger  - <vapier@gentoo.org>
+ * Copyright 2005-2012 Ned Ludd        - <solar@gentoo.org>
+ * Copyright 2005-2012 Mike Frysinger  - <vapier@gentoo.org>
  *
  * Make sure all of the common stuff is setup as we expect
  */
@@ -14,20 +13,30 @@
 
 #include "porting.h"
 #include "xfuncs.h"
+#include "security.h"
 
 #ifndef VERSION
-# define VERSION "cvs"
+# define VERSION "git"
+#endif
+#ifndef VCSID
+# define VCSID "<unknown>"
+#endif
+
+#ifdef EBUG
+# define USE_DEBUG 1
+#else
+# define USE_DEBUG 0
 #endif
 
 /* ELF love */
 #include "elf.h"
 #include "paxelf.h"
+#include "paxldso.h"
+
+/* Mach-O love */
+#include "macho.h"
+#include "paxmacho.h"
 
-/* MACH-O sucks */
-/*
- * #include "macho.h"
- * #include "paxmacho.h"
-*/
 extern char do_reverse_endian;
 
 #ifdef IN_paxinc
@@ -35,6 +44,8 @@ typedef struct {
 	int fd;
 	const char *filename;
 	size_t skip;
+	char *extfn;
+	bool verbose;
 } archive_handle;
 #else
 typedef void archive_handle;
@@ -61,10 +72,12 @@ typedef struct {
 	} buf;
 #endif
 } archive_member;
-archive_handle *ar_open_fd(const char *filename, int fd);
-archive_handle *ar_open(const char *filename);
+archive_handle *ar_open_fd(const char *filename, int fd, bool verbose);
+archive_handle *ar_open(const char *filename, bool verbose);
 archive_member *ar_next(archive_handle *);
 
+const char *strfileperms(const char *fname);
+
 /* Get a value 'X', compensating for endianness. */
 #define EGET(X) \
 	(__extension__ ({ \
@@ -74,7 +87,7 @@ archive_member *ar_next(archive_handle *);
 		} else if (sizeof(X) == 2) { __res = bswap_16((X)); \
 		} else if (sizeof(X) == 4) { __res = bswap_32((X)); \
 		} else if (sizeof(X) == 8) { __res = bswap_64((X)); \
-		} else { errf("EGET failed ;( (sizeof(X) == %i)", (int)sizeof(X)); } \
+		} else { errf("EGET failed :( (sizeof(X) == %i)", (int)sizeof(X)); } \
 		__res; \
 	}))
 
@@ -86,18 +99,21 @@ archive_member *ar_next(archive_handle *);
 		} else if (sizeof(Y) == 2) { Y = bswap_16((uint16_t)(X)); \
 		} else if (sizeof(Y) == 4) { Y = bswap_32((uint32_t)(X)); \
 		} else if (sizeof(Y) == 8) { Y = bswap_64((uint64_t)(X)); \
-		} else { errf("ESET failed ;( (size(Y) == %i)", (int)sizeof(Y)); } \
+		} else { errf("ESET failed :( (size(Y) == %i)", (int)sizeof(Y)); } \
 	} while (0)
 
+/* alignment helpers */
+#define ALIGN_DOWN(base, size) ((base) & -((__typeof__(base)) (size)))
+#define ALIGN_UP(base, size)   ALIGN_DOWN((base) + (size) - 1, (size))
+#define PTR_ALIGN_DOWN(base, size) ((__typeof__(base))ALIGN_DOWN((uintptr_t)(base), (size)))
+#define PTR_ALIGN_UP(base, size)   ((__typeof__(base))ALIGN_UP  ((uintptr_t)(base), (size)))
+
 /* helper functions for showing errors */
-#define color 1
-#define COLOR(c,b) (color ? "\e[" c ";" b "m" : "")
-#define NORM      COLOR("00", "00")
-#define RED       COLOR("31", "01")
-#define YELLOW    COLOR("33", "01")
+extern const char *NORM, *RED, *YELLOW;
+void color_init(bool disable);
 
 /* constant pointer to a constant buffer ... each program needs to set this */
-extern const char * const argv0;
+extern const char argv0[];
 
 /* we need the space before the last comma or we trigger a bug in gcc-2 :( */
 #define warn(fmt, args...) \
@@ -114,4 +130,9 @@ extern const char * const argv0;
 #define errf(fmt, args...) _err(warnf, fmt, ## args)
 #define errp(fmt, args...) _err(warnp, fmt , ## args)
 
+/* File system helper functions. */
+extern int root_fd;
+FILE *fopenat_r(int dir_fd, const char *path);
+const char *root_rel_path(const char *path);
+
 #endif /* _PAX_INC_H */
