@@ -12,8 +12,7 @@
  *  cc -o pspax pspax.c -DWANT_SYSCAP -lcap
  */
 
-static const char *rcsid = "$Id: pspax.c,v 1.38 2007/08/20 09:54:15 vapier Exp $";
-const char * const argv0 = "pspax";
+const char argv0[] = "pspax";
 
 #include "paxinc.h"
 #include <grp.h>
@@ -28,49 +27,104 @@ const char * const argv0 = "pspax";
 
 #define PROC_DIR "/proc"
 
-
-
 /* variables to control behavior */
 static char show_all = 0;
 static char verbose = 0;
 static char show_banner = 1;
 static char show_phdr = 0;
+static char show_addr = 0;
 static char noexec = 1;
 static char writeexec = 1;
-
+static char wide_output = 0;
 static pid_t show_pid = 0;
-static uid_t show_uid = -1;
-static gid_t show_gid = -1;
+static uid_t show_uid = (uid_t)-1;
+static gid_t show_gid = (gid_t)-1;
 
-static char *get_proc_name(pid_t pid)
+static int proc_open(int pfd, const char *file)
 {
-	FILE *fp;
-	static char str[__PAX_UTILS_PATH_MAX];
-	memset(&str, 0, sizeof(str));
+	return openat(pfd, file, O_RDONLY|O_CLOEXEC);
+}
 
-	snprintf(str, sizeof(str), PROC_DIR "/%u/stat", pid);
-	if ((fp = fopen(str, "r")) == NULL)
+static FILE *proc_fopen(int pfd, const char *file)
+{
+	int fd;
+	FILE *fp;
+
+	fd = proc_open(pfd, file);
+	if (fd == -1)
 		return NULL;
 
-	memset(&str, 0, sizeof(str));
+	fp = fdopen(fd, "re");
+	if (fp == NULL)
+		close(fd);
+
+	return fp;
+}
+
+static elfobj *proc_readelf(int pfd)
+{
+	int fd;
+	elfobj *elf;
+
+	fd = proc_open(pfd, "exe");
+	if (fd == -1)
+		return NULL;
+
+	elf = readelf_fd("proc/exe", fd, 0);
+	close(fd);
+	return elf;
+}
+
+static const char *get_proc_name_cmdline(int pfd)
+{
+	FILE *fp;
+	static char str[1024];
+
+	fp = proc_fopen(pfd, "cmdline");
+	if (fp == NULL)
+		return NULL;
+
+	if (fscanf(fp, "%s.1023", str) != 1) {
+		fclose(fp);
+		return NULL;
+	}
+	fclose(fp);
+
+	return (str);
+}
+
+static const char *get_proc_name(int pfd)
+{
+	FILE *fp;
+	static char str[BUFSIZ];
+
+	if (wide_output)
+		return get_proc_name_cmdline(pfd);
+
+	fp = proc_fopen(pfd, "stat");
+	if (fp == NULL)
+		return NULL;
+
+	if (fscanf(fp, "%*d %s.16", str) != 1) {
+		fclose(fp);
+		return NULL;
+	}
 
-	fscanf(fp, "%*d %s.16", str);
 	if (*str) {
 		str[strlen(str) - 1] = '\0';
 		str[16] = 0;
 	}
 	fclose(fp);
+
 	return (str+1);
 }
 
-static int get_proc_maps(pid_t pid)
+static int get_proc_maps(int pfd)
 {
-	static char str[__PAX_UTILS_PATH_MAX];
 	FILE *fp;
+	static char str[BUFSIZ];
 
-	snprintf(str, sizeof(str), PROC_DIR "/%u/maps", pid);
-
-	if ((fp = fopen(str, "r")) == NULL)
+	if ((fp = proc_fopen(pfd, "maps")) == NULL)
 		return -1;
 
 	while (fgets(str, sizeof(str), fp)) {
@@ -97,17 +151,16 @@ static int get_proc_maps(pid_t pid)
 		}
 	}
 	fclose(fp);
+
 	return 0;
 }
 
-static int print_executable_mappings(pid_t pid)
+static int print_executable_mappings(int pfd)
 {
-	static char str[__PAX_UTILS_PATH_MAX];
 	FILE *fp;
+	static char str[BUFSIZ];
 
-	snprintf(str, sizeof(str), PROC_DIR "/%u/maps", pid);
-
-	if ((fp = fopen(str, "r")) == NULL)
+	if ((fp = proc_fopen(pfd, "maps")) == NULL)
 		return -1;
 
 	while (fgets(str, sizeof(str), fp)) {
@@ -132,6 +185,7 @@ static int print_executable_mappings(pid_t pid)
 		}
 	}
 	fclose(fp);
+
 	return 0;
 }
 
@@ -144,28 +198,24 @@ static int print_executable_mappings(pid_t pid)
 # define NOTE_TO_SELF
 #endif
 
-static struct passwd *get_proc_passwd(pid_t pid)
+static const struct passwd *get_proc_passwd(int pfd)
 {
 	struct stat st;
-	struct passwd *pwd;
-	static char str[__PAX_UTILS_PATH_MAX];
+	const struct passwd *pwd = NULL;
 
-	snprintf(str, sizeof(str), PROC_DIR "/%u/stat", pid);
+	if (fstatat(pfd, "stat", &st, AT_SYMLINK_NOFOLLOW) != -1)
+		pwd = getpwuid(st.st_uid);
 
-	if ((stat(str, &st)) != (-1))
-		if ((pwd = getpwuid(st.st_uid)) != NULL)
-			return pwd;
-	return NULL;
+	return pwd;
 }
 
-static char *get_proc_status(pid_t pid, const char *name)
+static const char *get_proc_status(int pfd, const char *name)
 {
 	FILE *fp;
 	size_t len;
-	static char str[__PAX_UTILS_PATH_MAX];
+	static char str[BUFSIZ];
 
-	snprintf(str, sizeof(str), PROC_DIR "/%u/status", pid);
-	if ((fp = fopen(str, "r")) == NULL)
+	if ((fp = proc_fopen(pfd, "status")) == NULL)
 		return NULL;
 
 	len = strlen(name);
@@ -179,65 +229,78 @@ static char *get_proc_status(pid_t pid, const char *name)
 		}
 	}
 	fclose(fp);
+
 	return NULL;
 }
 
-static char *get_pid_attr(pid_t pid)
+static const char *get_pid_attr(int pfd)
 {
 	FILE *fp;
 	char *p;
-	char str[32];
 	static char buf[BUFSIZ];
 
-	memset(buf, 0, sizeof(buf));
-
-	snprintf(str, sizeof(str), PROC_DIR "/%u/attr/current", pid);
-	if ((fp = fopen(str, "r")) == NULL)
+	if ((fp = proc_fopen(pfd, "attr/current")) == NULL)
 		return NULL;
+
 	if (fgets(buf, sizeof(buf), fp) != NULL)
 		if ((p = strchr(buf, '\n')) != NULL)
 			*p = 0;
 	fclose(fp);
+
 	return buf;
 }
 
-static const char *get_proc_type(pid_t pid)
+static const char *get_pid_addr(int pfd)
 {
-	char fname[32];
-	elfobj *elf = NULL;
-	char *ret = NULL;
+	FILE *fp;
+	char *p;
+	static char buf[BUFSIZ];
 
-	snprintf(fname, sizeof(fname), PROC_DIR "/%u/exe", pid);
-	if ((elf = readelf(fname)) == NULL)
-		return ret;
-	ret = (char *)get_elfetype(elf);
+	if ((fp = proc_fopen(pfd, "ipaddr")) == NULL)
+		return NULL;
+
+	if (fgets(buf, sizeof(buf), fp) != NULL)
+		if ((p = strchr(buf, '\n')) != NULL)
+			*p = 0;
+	fclose(fp);
+
+	return buf;
+}
+
+static const char *get_proc_type(int pfd)
+{
+	elfobj *elf;
+	const char *ret;
+
+	elf = proc_readelf(pfd);
+	if (elf == NULL)
+		return NULL;
+
+	ret = get_elfetype(elf);
 	unreadelf(elf);
 	return ret;
 }
 
-static char *scanelf_file_phdr(elfobj *elf)
+static const char *scanelf_file_phdr(elfobj *elf)
 {
 	static char ret[8];
 	unsigned long i, off, multi_stack, multi_load;
-	int max_pt_load;
 
 	memcpy(ret, "--- ---\0", 8);
 
 	multi_stack = multi_load = 0;
-	max_pt_load = elf_max_pt_load(elf);
 
 	if (elf->phdr) {
 	uint32_t flags;
 #define SHOW_PHDR(B) \
 	if (elf->elf_class == ELFCLASS ## B) { \
-	Elf ## B ## _Ehdr *ehdr = EHDR ## B (elf->ehdr); \
-	Elf ## B ## _Phdr *phdr = PHDR ## B (elf->phdr); \
+	const Elf ## B ## _Ehdr *ehdr = EHDR ## B (elf->ehdr); \
+	const Elf ## B ## _Phdr *phdr = PHDR ## B (elf->phdr); \
 	for (i = 0; i < EGET(ehdr->e_phnum); i++) { \
 		if (EGET(phdr[i].p_type) == PT_GNU_STACK) { \
 			if (multi_stack++) warnf("%s: multiple PT_GNU_STACK's !?", elf->filename); \
 			off = 0; \
 		} else if (EGET(phdr[i].p_type) == PT_LOAD) { \
-			if (multi_load++ > max_pt_load) warnf("%s: more than %i PT_LOAD's !?", elf->filename, max_pt_load); \
 			off = 4; \
 		} else \
 			continue; \
@@ -252,145 +315,167 @@ static char *scanelf_file_phdr(elfobj *elf)
 	return ret;
 }
 /* we scan the elf file two times when the -e flag is given. But we don't need -e very often so big deal */
-static const char *get_proc_phdr(pid_t pid)
+static const char *get_proc_phdr(int pfd)
 {
-	char fname[32];
-	elfobj *elf = NULL;
-	char *ret = NULL;
+	elfobj *elf;
+	const char *ret;
 
-	snprintf(fname, sizeof(fname), PROC_DIR "/%u/exe", pid);
-	if ((elf = readelf(fname)) == NULL)
-		return ret;
-	ret = (char *) scanelf_file_phdr(elf);
+	elf = proc_readelf(pfd);
+	if (elf == NULL)
+		return NULL;
+
+	ret = scanelf_file_phdr(elf);
 	unreadelf(elf);
 	return ret;
 }
 
-
 static void pspax(const char *find_name)
 {
 	register DIR *dir;
 	register struct dirent *de;
 	pid_t pid;
 	pid_t ppid = show_pid;
-	int have_attr, wx;
-	struct passwd *pwd;
-	struct stat st;
-	const char *pax, *type, *name, *caps, *attr;
-	WRAP_SYSCAP(ssize_t length; cap_t cap_d;);
+	int have_attr, have_addr, wx;
+	const struct passwd *pwd;
+	const char *pax, *type, *name, *attr, *addr;
+	char *caps;
+	int pfd;
+	WRAP_SYSCAP(ssize_t length; cap_t cap_d;)
 
 	WRAP_SYSCAP(cap_d = cap_init());
 
-	caps = NULL;
+	dir = opendir(PROC_DIR);
+	if (dir == NULL || chdir(PROC_DIR))
+		errp(PROC_DIR);
 
-	chdir(PROC_DIR);
-	if (!(dir = opendir(PROC_DIR))) {
-		perror(PROC_DIR);
-		exit(EXIT_FAILURE);
-	}
-
-	if (access("/proc/self/attr/current", R_OK) != (-1))
+	if (access("/proc/self/attr/current", R_OK) != -1)
 		have_attr = 1;
 	else
 		have_attr = 0;
 
+	if ((access("/proc/self/ipaddr", R_OK) != -1) && show_addr)
+		have_addr = 1;
+	else
+		have_addr = 0;
+
 	if (show_banner)
-		printf("%-8s %-6s %-6s %-4s %-10s %-16s %-4s %-4s %s\n",
-		       "USER", "PID", "PAX", "MAPS", "ETYPE", "NAME", "CAPS", "ATTR", show_phdr ? "STACK LOAD" : "");
+		printf("%-8s %-6s %-6s %-4s %-10s %-16s %-4s %-4s %s %s\n",
+		       "USER", "PID", "PAX", "MAPS", "ETYPE", "NAME", "CAPS", have_attr ? "ATTR" : "",
+			have_addr ? "IPADDR" : "", show_phdr ? "STACK LOAD" : "");
 
 	while ((de = readdir(dir))) {
-		errno = 0;
-		stat(de->d_name, &st);
-		if ((errno != ENOENT) && (errno != EACCES)) {
-			pid = (pid_t) atoi((char *) basename((char *) de->d_name));
-			if (find_name && pid) {
-				char *str = get_proc_name(pid);
-	                        if (!str) continue;
-				if (strcmp(str, find_name) != 0)
-					pid = 0;
-			}
-			if (((ppid > 0) && (pid != ppid)) || (!pid))
-				continue;
-
-			wx = get_proc_maps(pid);
-
-			if (noexec != writeexec) {
-				if ((wx == 1) && (writeexec != wx))
-					goto next_pid;
-
-				if ((wx == 0) && (writeexec))
-					goto next_pid;
-			}
-
-			pwd  = get_proc_passwd(pid);
-			pax  = get_proc_status(pid, "PAX");
-			type = get_proc_type(pid);
-			name = get_proc_name(pid);
-			attr = (have_attr ? get_pid_attr(pid) : NULL);
-
-			if (show_uid != (-1) && pwd)
-				if (pwd->pw_uid != show_uid)
-					continue;
-
-			if (show_gid != (-1) && pwd)
-				if (pwd->pw_gid != show_gid)
-					continue;
-
-			/* this is a non-POSIX function */
-			WRAP_SYSCAP(capgetp(pid, cap_d));
-			WRAP_SYSCAP(caps = cap_to_text(cap_d, &length));
-
-			if (show_all || type) {
-				printf("%-8s %-6d %-6s %-4s %-10s %-16s %-4s %s %s\n",
-				       pwd  ? pwd->pw_name : "--------",
-				       pid,
-				       pax  ? pax  : "---",
-				       (wx == 1) ? "w|x" : (wx == -1) ? "---" : "w^x",
-				       type ? type : "-------",
-				       name ? name : "-----",
-				       caps ? caps : " = ",
-				       attr ? attr : "-", show_phdr ? get_proc_phdr(pid) : "");
-				if (verbose && wx)
-					print_executable_mappings(pid);
-			}
-
-			WRAP_SYSCAP(if (caps) cap_free((void *)caps));
-
-		next_pid:
+		/* Check the name first if it's an int as it's faster. */
+		pid = atoi(de->d_name);
+		if (pid == 0)
 			continue;
+
+		/* Get an open handle so the kernel won't reap on us later. */
+		pfd = open(de->d_name, O_RDONLY|O_CLOEXEC|O_PATH|O_DIRECTORY);
+		if (pfd == -1)
+			continue;
+
+		if (find_name && pid) {
+			const char *str = get_proc_name(pfd);
+			if (!str)
+				goto next_pid;
+			if (strcmp(str, find_name) != 0)
+				pid = 0;
 		}
+		if (ppid > 0 && pid != ppid)
+			goto next_pid;
+
+		wx = get_proc_maps(pfd);
+
+		if (noexec != writeexec) {
+			if (wx == 1 && writeexec != wx)
+				goto next_pid;
+
+			if (wx == 0 && writeexec)
+				goto next_pid;
+		}
+
+		pwd  = get_proc_passwd(pfd);
+		pax  = get_proc_status(pfd, "PAX");
+		type = get_proc_type(pfd);
+		name = get_proc_name(pfd);
+		attr = (have_attr ? get_pid_attr(pfd) : NULL);
+		addr = (have_addr ? get_pid_addr(pfd) : NULL);
+
+		if (pwd) {
+			if (show_uid != (uid_t)-1)
+				if (pwd->pw_uid != show_uid)
+					goto next_pid;
+
+			if (show_gid != (gid_t)-1)
+				if (pwd->pw_gid != show_gid)
+					goto next_pid;
+		}
+
+		/* this is a non-POSIX function */
+		caps = NULL;
+		WRAP_SYSCAP(capgetp(pfd, cap_d));
+		WRAP_SYSCAP(caps = cap_to_text(cap_d, &length));
+
+		if (pwd && strlen(pwd->pw_name) >= 8)
+			pwd->pw_name[8] = 0;
+
+		if (show_all || type) {
+			printf("%-8s %-6d %-6s %-4s %-10s %-16s %-4s %s %s %s\n",
+			       pwd  ? pwd->pw_name : "--------",
+			       pid,
+			       pax  ? pax  : "---",
+			       (wx == 1) ? "w|x" : (wx == -1) ? "---" : "w^x",
+			       type ? type : "-------",
+			       name ? name : "-----",
+			       caps ? caps : " = ",
+			       attr ? attr : "",
+			       addr ? addr : "",
+			       show_phdr ? get_proc_phdr(pfd) : "");
+			if (verbose && wx)
+				print_executable_mappings(pfd);
+		}
+
+		WRAP_SYSCAP(if (caps) cap_free(caps));
+
+ next_pid:
+		close(pfd);
 	}
 	closedir(dir);
 }
 
-
-
 /* usage / invocation handling functions */
-#define PARSE_FLAGS "aep:u:g:nwvBhV"
+#define PARSE_FLAGS "aeip:u:g:nwWvCBhV"
 #define a_argument required_argument
 static struct option const long_opts[] = {
 	{"all",       no_argument, NULL, 'a'},
 	{"header",    no_argument, NULL, 'e'},
+	{"ipaddr",    no_argument, NULL, 'i'},
 	{"pid",        a_argument, NULL, 'p'},
 	{"user",       a_argument, NULL, 'u'},
 	{"group",      a_argument, NULL, 'g'},
 	{"nx",        no_argument, NULL, 'n'},
 	{"wx",        no_argument, NULL, 'w'},
+	{"wide",      no_argument, NULL, 'W'},
 	{"verbose",   no_argument, NULL, 'v'},
+	{"nocolor",   no_argument, NULL, 'C'},
 	{"nobanner",  no_argument, NULL, 'B'},
 	{"help",      no_argument, NULL, 'h'},
 	{"version",   no_argument, NULL, 'V'},
 	{NULL,        no_argument, NULL, 0x0}
 };
-static const char *opts_help[] = {
+
+static const char * const opts_help[] = {
 	"Show all processes",
 	"Print GNU_STACK/PT_LOAD markings",
+	"Print ipaddr info if supported",
 	"Process ID/pid #",
 	"Process user/uid #",
 	"Process group/gid #",
 	"Only display w^x processes",
 	"Only display w|x processes",
+	"Wide output display of cmdline",
 	"Be verbose about executable mappings",
+	"Don't emit color in output",
 	"Don't display the header",
 	"Print this help and exit",
 	"Print version and exit",
@@ -419,27 +504,30 @@ static void usage(int status)
 static void parseargs(int argc, char *argv[])
 {
 	int flag;
-	struct passwd *pwd = NULL;
-	struct  group *gwd = NULL;
+	const struct passwd *pwd = NULL;
+	const struct  group *gwd = NULL;
 
 	opterr = 0;
 	while ((flag=getopt_long(argc, argv, PARSE_FLAGS, long_opts, NULL)) != -1) {
 		switch (flag) {
 
 		case 'V':                        /* version info */
-			printf("pax-utils-%s: %s compiled %s\n%s\n"
+			printf("pax-utils-%s: %s\n"
 			       "%s written for Gentoo by <solar and vapier @ gentoo.org>\n",
-			       VERSION, __FILE__, __DATE__, rcsid, argv0);
+			       VERSION, VCSID, argv0);
 			exit(EXIT_SUCCESS);
 			break;
 		case 'h': usage(EXIT_SUCCESS); break;
 
+		case 'C': color_init(true); break;
 		case 'B': show_banner = 0; break;
 		case 'a': show_all = 1; break;
 		case 'e': show_phdr = 1; break;
+		case 'i': show_addr = 1; break;
 		case 'p': show_pid = atoi(optarg); break;
 		case 'n': noexec = 1; writeexec = 0; break;
 		case 'w': noexec = 0; writeexec = 1; break;
+		case 'W': wide_output = 1; break;
 		case 'v': verbose++; break;
 		case 'u':
 			show_uid = atoi(optarg);
@@ -473,12 +561,15 @@ static void parseargs(int argc, char *argv[])
 	}
 }
 
-
-
 int main(int argc, char *argv[])
 {
-	char *name = NULL;
+	const char *name = NULL;
 
+	/* We unshare pidns but don't actually enter it.  That means
+	 * we still get to scan /proc, but just not fork children.  */
+	security_init(false);
+
+	color_init(false);
 	parseargs(argc, argv);
 
 	if ((optind < argc) && (show_pid == 0))
